//
//  main.m
//  ICNS Parser
//
//  Created by Keith Duncan on 16/08/2012.
//  Copyright (c) 2012 Keith Duncan. All rights reserved.
//

#import <Foundation/Foundation.h>
#import <CoreFoundation/CoreFoundation.h>

@interface NSDictionary (Additions)
- (id)objectForKeyedSubscript:(id)key;
- (void)setObject:(id)object forKeyedSubscript:(id)key;
@end
@implementation NSDictionary (Additions)
- (id)objectForKeyedSubscript:(id)key {
	return [self objectForKey:key];
}
- (void)setObject:(id)object forKeyedSubscript:(id)key {
	object ? [(id)self setObject:object forKey:key] : [(id)self removeObjectForKey:key];
}
@end

static void Print(char const *string) {
	fprintf(stdout, "%s\n", string);
}

static NSString *FourCCString(FourCharCode c) {
    char fourcc[sizeof("0xFFFF")] = {0};
    int i;
    
	//not a fourcc or twocc
	if (c < '\0\0AA') {
		snprintf(fourcc, sizeof(fourcc), "%#x", (unsigned)c);
	}
	else {
		for (i = 0; i < 4; i++) {
			fourcc[i] = c >> 8*(3-i);
		}
		fourcc[4] = '\0';
	}
    
    return [NSString stringWithCString:fourcc encoding:NSASCIIStringEncoding];
}

static FourCharCode StringFourCC(NSString *string) {
	NSCParameterAssert([string length] == 4);
	
	char const *cString = [string cStringUsingEncoding:NSASCIIStringEncoding];
	OSType type = (cString[0] << 24) | (cString[1] << 16) | (cString[2] << 8) | cString[3];
	return type;
}

static NSArray *IcnsTypeArray(void) {
	NSArray *types = @[
		@{ @"type" : @"ICNS", @"name" : @"Header", @"min_os" : @"10.7", @"encoding" : @"Custom" },
		@{ @"type" : @"TOC ", @"name" : @"Table of Contents", @"min_os" : @"10.7", @"encoding" : @"Custom" },
		
		@{ @"type" : @"ics#", @"name" : @"16×16 1-bit mask", @"min_os" : @"6.0", @"encoding" : @"" },
		@{ @"type" : @"ics4", @"name" : @"16×16 4-bit icon", @"min_os" : @"7.0", @"encoding" : @"" },
		@{ @"type" : @"ics8", @"name" : @"16×16 8-bit icon", @"min_os" : @"7.0", @"encoding" : @"" },
		@{ @"type" : @"is32", @"name" : @"16×16 24-bit icon", @"min_os" : @"8.5", @"encoding" : @"rle" },
		@{ @"type" : @"s8mk", @"name" : @"16x16 8-bit mask", @"min_os" : @"8.5", @"encoding" : @"" },
		
		@{ @"type" : @"ICN#", @"name" : @"32×32 1-bit mono icon with 1-bit mask", @"min_os" : @"6.0", @"encoding" : @"" },
		@{ @"type" : @"icl4", @"name" : @"32×32 4-bit mask", @"min_os" : @"7.0", @"encoding" : @"" },
		@{ @"type" : @"icl8", @"name" : @"32×32 8-bit mask", @"min_os" : @"7.0", @"encoding" : @"" },
		@{ @"type" : @"il32", @"name" : @"32x32 24-bit icon", @"min_os" : @"8.5", @"encoding" : @"rle" },
		@{ @"type" : @"l8mk", @"name" : @"32×32 8-bit mask", @"min_os" : @"8.5", @"encoding" : @"" },
		
		@{ @"type" : @"ich#", @"name" : @"48×48 1-bit mask", @"min_os" : @"8.5", @"encoding" : @"" },
		@{ @"type" : @"ich4", @"name" : @"48×48 4-bit icon", @"min_os" : @"8.5", @"encoding" : @"" },
		@{ @"type" : @"ich8", @"name" : @"48×48 8-bit icon", @"min_os" : @"8.5", @"encoding" : @"" },
		@{ @"type" : @"ih32", @"name" : @"48×48 24-bit icon", @"min_os" : @"8.5", @"encoding" : @"rle" },
		@{ @"type" : @"h8mk", @"name" : @"48×48 8-bit mask", @"min_os" : @"8.5", @"encoding" : @"" },
		
		@{ @"type" : @"it32", @"name" : @"128×128 24-bit icon", @"min_os" : @"10.0", @"encoding" : @"rle" },
		@{ @"type" : @"t8mk", @"name" : @"128×128 8-bit mask", @"min_os" : @"10.0", @"encoding" : @"" },
		
		@{ @"type" : @"icp4", @"name" : @"16x16", @"min_os" : @"10.7", @"encoding" : @"png" },
		@{ @"type" : @"icp5", @"name" : @"32x32", @"min_os" : @"10.7", @"encoding" : @"png" },
		@{ @"type" : @"ic07", @"name" : @"128x128", @"min_os" : @"10.7", @"encoding" : @"png" },
		@{ @"type" : @"ic08", @"name" : @"256×256", @"min_os" : @"10.5", @"encoding" : @"png" },
		@{ @"type" : @"ic09", @"name" : @"512×512", @"min_os" : @"10.5", @"encoding" : @"png" },
		@{ @"type" : @"ic10", @"name" : @"512x512@2x", @"min_os" : @"10.7", @"encoding" : @"png" },
		
		@{ @"type" : @"ic11", @"name" : @"16x16@2x", @"min_os" : @"10.8", @"encoding" : @"png" },
		@{ @"type" : @"ic12", @"name" : @"32x32@2x", @"min_os" : @"10.8", @"encoding" : @"png" },
		@{ @"type" : @"ic13", @"name" : @"128x128@2x", @"min_os" : @"10.8", @"encoding" : @"png" },
		@{ @"type" : @"ic14", @"name" : @"256x256@2x", @"min_os" : @"10.8", @"encoding" : @"png" },
		
		@{ @"type" : @"icnV", @"name" : @"Icon Composer Version", @"encoding" : @"Custom" },
	];
	return types;
}

static id IcnsTypeFrom(NSString *type, NSArray *types) {
	NSPredicate *predicate = [NSPredicate predicateWithFormat:@"type == %@", type];
	NSArray *filteredTypes = [types filteredArrayUsingPredicate:predicate];
	return [filteredTypes lastObject];
}

static id IcnsType(NSString *type) {
	return IcnsTypeFrom(type, IcnsTypeArray());
}

static uint8_t const IcnsHeader[4] = {
	0x69,
	0x63,
	0x6e,
	0x73,
};

static NSArray *ParseIconRegionsForICNSData(NSData *icnsData, NSError **errorRef) {
	NSMutableArray *types = [NSMutableArray array];
	
	if (errorRef != NULL) {
		NSDictionary *errorInfo = @{
			NSLocalizedDescriptionKey : @"icns parse error",
		};
		*errorRef = [NSError errorWithDomain:@"com.keith-duncan.icns_read" code:0 userInfo:errorInfo];
	}
	
	void const *icnsBytes = [icnsData bytes];
	size_t icnsLength = [icnsData length];
	
	void const *icnsBytesEnd = (icnsBytes + icnsLength);
	uintptr_t icnsCursor = 0;
	
	// ICNS Header
	{
		void const *headerBytes = (icnsBytes + icnsCursor);
		uintptr_t headerCursor = 0;
		
		void const *dataHeader = (headerBytes + headerCursor);
		size_t dataHeaderSize = 4;
		
		if ((dataHeader + (headerCursor + dataHeaderSize)) > icnsBytesEnd) {
			return nil;
		}
		headerCursor += dataHeaderSize;
		
		if (memcmp(IcnsHeader, (uint8_t *)dataHeader, dataHeaderSize) != 0) {
			return nil;
		}
		
		void const *fileLengthRef = (headerBytes + headerCursor);
		size_t fileLengthSize = 4;
		
		if ((fileLengthRef + (headerCursor + fileLengthSize)) > icnsBytesEnd) {
			return nil;
		}
		headerCursor += fileLengthSize;
		
		uint32_t fileLength = CFSwapInt32BigToHost(*(uint32_t *)fileLengthRef);
		if (fileLength != icnsLength) {
			return nil;
		}
		
		NSData *icnsHeaderData = [NSData dataWithBytes:headerBytes length:headerCursor];
		[types addObject:@{ @"type" : @"ICNS", @"typeOffset" : @(icnsCursor), @"typeLength" : @(icnsLength), @"typeData" : icnsHeaderData }];
		
		icnsCursor += headerCursor;
	}
	
	while (icnsCursor < icnsLength) {
		// Icon Header
		void const *iconBytes = (icnsBytes + icnsCursor);
		uintptr_t iconCursor = 0;
		
		uint32_t const *iconType = (iconBytes + iconCursor);
		size_t iconTypeSize = 4;
		
		if ((iconBytes + (iconCursor + iconTypeSize)) > icnsBytesEnd) {
			return nil;
		}
		iconCursor += iconTypeSize;
		NSString *iconTypeString = FourCCString((OSType)CFSwapInt32BigToHost(*iconType));
		
		uint32_t const *iconLengthRef = (iconBytes + iconCursor);
		size_t iconLengthSize = 4;
		
		if ((iconBytes + (iconCursor + iconLengthSize)) > icnsBytesEnd) {
			return nil;
		}
		iconCursor += iconLengthSize;
		
		uint32_t iconLength = CFSwapInt32BigToHost(*iconLengthRef);
		if ((iconBytes + iconLength) > icnsBytesEnd) {
			return nil;
		}
		
		void const *iconBitmap = (iconBytes + iconCursor);
		size_t iconBitmapSize = (iconLength - iconLengthSize - iconTypeSize);
		NSData *iconBitmapData __attribute__((unused)) = [NSData dataWithBytes:iconBitmap length:iconBitmapSize];
		iconCursor += iconBitmapSize;
		
		NSData *typeData = [NSData dataWithBytes:iconBytes length:iconCursor];
		NSData *bodyData = [NSData dataWithBytes:iconBitmap length:iconBitmapSize];
		
		[types addObject:@{ @"type" : iconTypeString, @"typeOffset" : @(icnsCursor), @"typeLength" : @([typeData length]), @"typeData" : typeData, @"bodyData" : bodyData }];
		
		icnsCursor += iconCursor;
	}
	
	return types;
}

static void PrintICNSAtPath(char const *pathBytes) {
	NSMutableDictionary *topLevelObject = [NSMutableDictionary dictionary];
	
	NSString *path = [[NSFileManager defaultManager] stringWithFileSystemRepresentation:pathBytes length:strlen(pathBytes)];
	topLevelObject[@"path"] = path;
	
	do {
		NSError *icnsDataError = nil;
		NSData *icnsData = [NSData dataWithContentsOfFile:path options:(NSDataReadingOptions)0 error:&icnsDataError];
		if (icnsData == nil) {
			topLevelObject[@"error"] = icnsDataError;
			break;
		}
		
		NSError *icnsPartsError = nil;
		NSArray *icnsParts = ParseIconRegionsForICNSData(icnsData, &icnsPartsError);
		if (icnsParts == nil) {
			NSMutableDictionary *errorDescription = [NSMutableDictionary dictionary];
			errorDescription[NSLocalizedDescriptionKey] = [icnsPartsError localizedDescription];
			errorDescription[NSLocalizedRecoverySuggestionErrorKey] = [icnsPartsError localizedRecoverySuggestion];
			topLevelObject[@"error"] = errorDescription;
			break;
		}
		
		NSArray *preferredOrder = IcnsTypeArray();
		
		NSPointerArray *accumulationArray = [NSPointerArray strongObjectsPointerArray];
		[accumulationArray setCount:[preferredOrder count]];
		
		NSMutableArray *types = [NSMutableArray array];
		[icnsParts enumerateObjectsUsingBlock:^ (id obj, NSUInteger idx, BOOL *stop) {
			NSString *type = obj[@"type"];
			
			NSUInteger preferredIndex = [preferredOrder indexOfObjectPassingTest:^ BOOL (id obj, NSUInteger idx, BOOL *stop) {
				return [obj[@"type"] isEqual:type];
			}];
			if (preferredIndex != NSNotFound) {
				[accumulationArray replacePointerAtIndex:preferredIndex withPointer:obj];
			}
			else {
				[accumulationArray addPointer:obj];
			}
			
			id typeDescription = IcnsType(type);
			
			NSMutableDictionary *currentTypeDictionary = [NSMutableDictionary dictionary];
			if (typeDescription != nil) {
				[currentTypeDictionary addEntriesFromDictionary:typeDescription];
				currentTypeDictionary[@"typeLength"] = obj[@"typeLength"];
			}
			else {
				currentTypeDictionary[@"type"] = type;
			}
			
			[types addObject:currentTypeDictionary];
		}];
		topLevelObject[@"types"] = types;
	} while (0);
	
	NSString *serialisedTopLevelObject = [[[NSString alloc] initWithData:[NSJSONSerialization dataWithJSONObject:topLevelObject options:NSJSONWritingPrettyPrinted error:NULL] encoding:NSUTF8StringEncoding] autorelease];
	Print([serialisedTopLevelObject UTF8String]);
}

id (^partWithType)(NSString *, NSData *) = ^ id (NSString *typeString, NSData *bodyData) {
	size_t typeSize = 4;
	uint32_t type = StringFourCC(typeString);
	
	size_t typeDataLengthSize = 4;
	uint32_t typeDataLength = (uint32_t)([bodyData length] + typeSize + typeDataLengthSize);
	
	size_t typeDataSize = (typeSize + typeDataLengthSize);
	void *typeData = malloc(typeDataSize);
	uintptr_t typeCursor = 0;
	
	void *typeRef = (typeData + typeCursor);
	typeCursor += typeSize;
	*(uint32_t *)typeRef = CFSwapInt32HostToBig(type);
	
	void *typeLengthRef = (typeData + typeCursor);
	typeCursor += typeDataLengthSize;
	*(uint32_t *)typeLengthRef = CFSwapInt32HostToBig(typeDataLength);
	
	NSMutableData *encodedTypeData = [NSMutableData dataWithBytes:typeData length:typeDataSize];
	[encodedTypeData appendData:bodyData];
	
	free(typeData);
	
	id part = @{ @"type" : typeString, @"typeLength" : @(typeDataLength), @"typeData" : encodedTypeData, @"bodyData" : bodyData };
	return part;
};

static int WritePartsToOutputPath(NSArray *parts, BOOL includeTableOfContents, NSString *outputPath, NSError **errorRef) {
	NSMutableArray *partsToWrite = [NSMutableArray arrayWithArray:parts];
	
	if (includeTableOfContents) {
		/*
			'TOC ' encoding
			
			{
				uint32_t type; // OSType, big endian
				uint32_t typeLength; // length of the whole 'part' encoded not just the bitmap
			}
		 */
		
		size_t bodyDataLength = sizeof(uint32_t) * ([parts count] * 2);
		void *bodyData = alloca(bodyDataLength);
		uintptr_t bodyCursor = 0;
		
		for (id currentIconPart in parts) {
			NSString *type = currentIconPart[@"type"];
			uint32_t typeValue = StringFourCC(type);
			void *typeValueRef = (bodyData + bodyCursor);
			*(uint32_t *)typeValueRef = CFSwapInt32HostToBig(typeValue);
			bodyCursor += 4;
			
			uint32_t lengthValue = (uint32_t)[currentIconPart[@"typeData"] length];
			void *lengthValueRef = (bodyData + bodyCursor);
			*(uint32_t *)lengthValueRef = CFSwapInt32HostToBig(lengthValue);
			bodyCursor += 4;
		}
		
		NSData *encodedBodyData = [NSData dataWithBytes:bodyData length:bodyDataLength];
		
		id part = partWithType(@"TOC ", encodedBodyData);
		[partsToWrite insertObject:part atIndex:0];
	}
	
	NSMutableData *bodyData = [NSMutableData data];
	for (id part in partsToWrite) {
		NSData *partData = part[@"typeData"];
		[bodyData appendData:partData];
	}
	
	id part = partWithType(@"ICNS", bodyData);
	NSMutableData *typeData = [NSMutableData dataWithData:part[@"typeData"]];
	[typeData replaceBytesInRange:NSMakeRange(0, 4) withBytes:IcnsHeader];
	
	NSError *writeError = nil;
	BOOL write = [typeData writeToFile:outputPath options:(NSDataWritingOptions)0 error:&writeError];
	if (!write) {
		return -1;
	}
	
	return 0;
}

static int ConvertIconSetAtPathToICNSAtPath(char const *inpath, char const *outpath) {
	NSString *inputPath = [[NSFileManager defaultManager] stringWithFileSystemRepresentation:inpath length:strlen(inpath)];
	NSString *outputPath = [[NSFileManager defaultManager] stringWithFileSystemRepresentation:outpath length:strlen(outpath)];
	
	if (![[inputPath pathExtension] isEqualToString:@"iconset"]) {
		Print("input path must be an iconset folder");
		return -1;
	}
	
	typedef NS_ENUM(NSUInteger, DeploymentTarget) {
		TenFive,
		TenSix,
		TenSeven,
		TenEight,
	};
	
	DeploymentTarget minimumDeploymentTarget = TenFive;
	DeploymentTarget maximumDeploymentTarget = TenEight;
	
	NSMutableArray *typesToInclude = [NSMutableArray array];
	
	void (^addTypes)(NSArray *types) = ^ void (NSArray *types) {
		[types enumerateObjectsUsingBlock:^(id obj, NSUInteger idx, BOOL *stop) {
			if ([typesToInclude containsObject:obj]) {
				return;
			}
			[typesToInclude addObject:obj];
		}];
	};
	
	BOOL (^deploymentTargetInRange)(DeploymentTarget deploymentTarget) = ^ BOOL (DeploymentTarget deploymentTarget) {
		return (minimumDeploymentTarget <= deploymentTarget && maximumDeploymentTarget >= deploymentTarget);
	};
	
	NSArray *tenFiveTypes = @[ @[ @"is32", @"s8mk" ], @[ @"il32", @"l8mk" ], @[ @"it32", @"t8mk" ], @"ic09" ];
	if (deploymentTargetInRange(TenFive)) {
		addTypes(tenFiveTypes);
	}
	NSArray *tenSixTypes = tenFiveTypes;
	if (deploymentTargetInRange(TenSix)) {
		addTypes(tenSixTypes);
	}
	NSArray *tenSevenTypes = @[ @"icp4", @"icp5", @"ic07", @"ic08", @"ic09", @"ic10" ];
	if (deploymentTargetInRange(TenSeven)) {
		addTypes(tenSevenTypes);
	}
	NSArray *tenEightTypes = @[ @"icp4", @"ic11", @"icp5", @"ic12", @"ic07", @"ic13", @"ic08", @"ic14", @"ic09", @"ic10" ];
	if (deploymentTargetInRange(TenEight)) {
		addTypes(tenEightTypes);
	}
	
	/*
		17 unique types to encode
		
		three encoders
		- rle
		- mask
		- png
	 */
	
	NSMutableArray *icnsParts = [NSMutableArray array];
	
	void (^encodeType)(NSString *type) = ^ void (NSString *type) {
		id typeDescription = IcnsType(type);
		
		NSString *description = typeDescription[@"description"];
		NSRange descriptionSpaceRange = [description rangeOfString:@" "];
		if (descriptionSpaceRange.location != NSNotFound) {
			description = [description substringToIndex:descriptionSpaceRange.location];
		}
		
		NSString *sourceFileName = [NSString stringWithFormat:@"icon_%@.png", description];
		NSString *sourceFilePath = [inputPath stringByAppendingPathComponent:sourceFileName];
		
		NSError *sourceDataError = nil;
		NSData *sourceData = [NSData dataWithContentsOfFile:sourceFilePath options:(NSDataReadingOptions)0 error:&sourceDataError];
		if (sourceData == nil) {
			Print([[NSString stringWithFormat:@"couldn't find source file named %@", sourceFileName] UTF8String]);
			exit(-1);
			return;
		}
		
		NSData *bodyData = nil;
		
		NSString *encoding = typeDescription[@"encoding"];
		NSString *pixelSize = [description substringToIndex:[description rangeOfString:@"x"].location];
		
		if ([encoding isEqualToString:@"png"]) {
			
		}
		else if ([encoding isEqualToString:@"png"]) {
			
		}
		
		[icnsParts addObject:partWithType(type, bodyData)];
	};
	
	for (id currentType in typesToInclude) {
		if ([currentType isKindOfClass:[NSString class]]) {
			encodeType(currentType);
		}
		else if ([currentType isKindOfClass:[NSArray class]]) {
			[currentType enumerateObjectsUsingBlock:^ (id obj, NSUInteger idx, BOOL *stop) { encodeType(obj); }];
		}
		else {
			Print([[NSString stringWithFormat:@"cannot encode type %@", currentType] UTF8String]);
			return -1;
		}
	}
	
	BOOL includeTableOfContents = (maximumDeploymentTarget >= TenSeven);
	return WritePartsToOutputPath(icnsParts, includeTableOfContents, outputPath, NULL);
}

static int CombineICNS(char const *inpath, char const *inpath2, char const *outpath) {
	NSString *inputPath = [[NSFileManager defaultManager] stringWithFileSystemRepresentation:inpath length:strlen(inpath)];
	NSArray *inputPathParts = ParseIconRegionsForICNSData([NSData dataWithContentsOfFile:inputPath], NULL);
	if (inputPathParts == nil) {
		return -1;
	}
	
	NSString *inputPath2 = [[NSFileManager defaultManager] stringWithFileSystemRepresentation:inpath2 length:strlen(inpath2)];
	NSArray *inputPath2Parts = ParseIconRegionsForICNSData([NSData dataWithContentsOfFile:inputPath2], NULL);
	if (inputPath2Parts == nil) {
		return -1;
	}
	
	NSMutableArray *combinedParts = [NSMutableArray array];
	
	void (^addTypeNotPresent)(NSArray *) = ^ void (NSArray *types) {
		[types enumerateObjectsUsingBlock:^ (id obj, NSUInteger idx, BOOL *stop) {
			NSString *type = obj[@"type"];
			if ([type isEqualToString:@"ICNS"]) {
				return;
			}
			if ([type isEqualToString:@"TOC "]) {
				return;
			}
			if ([type isEqualToString:@"icnV"]) {
				return;
			}
			
			NSPredicate *predicate = [NSPredicate predicateWithFormat:@"type == %@", type];
			if ([[combinedParts filteredArrayUsingPredicate:predicate] count] > 0) {
				return;
			}
			
			[combinedParts addObject:obj];
		}];
	};
	addTypeNotPresent(inputPathParts);
	addTypeNotPresent(inputPath2Parts);
	
	NSString *outputPath = [[NSFileManager defaultManager] stringWithFileSystemRepresentation:outpath length:strlen(outpath)];
	return WritePartsToOutputPath(combinedParts, YES, outputPath, NULL);
}

int main(int argc, char const **argv) {
	@autoreleasepool {
		if (argc <= 1) {
			Print([[NSString stringWithFormat:@"Usage: %s [read inpath] [write inpath outpath] [combine inpath1 inpath2 outpath]\n", argv[0]] UTF8String]);
			return -1;
		}
		char const *command = argv[1];
		
		if (strcmp(command, "read") == 0) {
			if (argc <= 2) {
				Print("no input path given");
				return -1;
			}
			
			for (int idx = 2; idx < argc; idx++) {
				char const *currentPath = argv[idx];
				PrintICNSAtPath(currentPath);
			}
			return 0;
		}
		
		if (strcmp(command, "write") == 0) {
			if (argc <= 2) {
				Print("no input path given");
			}
			char const *inpath = argv[2];
			
			if (argc <= 3) {
				Print("no output path given");
			}
			char const *outpath = argv[3];
			
			return ConvertIconSetAtPathToICNSAtPath(inpath, outpath);
		}
		
		if (strcmp(command, "combine") == 0) {
			if (argc <= 2) {
				Print("no input path given");
			}
			char const *inpath = argv[2];
			
			if (argc <= 3) {
				Print("no input path given");
			}
			char const *inpath2 = argv[3];
			
			if (argc <= 4) {
				Print("no output path given");
			}
			char const *outpath = argv[4];
			
			return CombineICNS(inpath, inpath2, outpath);
		}
		
		Print([[NSString stringWithFormat:@"unknown command %s", command] UTF8String]);
		return -1;
	}
}
